@@ -12,8 +12,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from __future__ import print_function
-
 import getopt
 import os
 import re
@@ -27,11 +25,14 @@ import ovs.ovsuuid
 import ovs.poller
 import ovs.stream
 import ovs.util
+import ovs.vlog
 from ovs.db import data
 from ovs.db import error
 from ovs.fatal_signal import signal_alarm
 
-import six
+vlog = ovs.vlog.Vlog("test-ovsdb")
+vlog.set_levels_from_string("console:dbg")
+vlog.init(None)
 
 
 def unbox_json(json):
@@ -162,7 +163,7 @@ def get_simple_printable_row_string(row, columns):
             s += "%s=%s " % (column, value)
     s = s.strip()
     s = re.sub('""|,|u?\'', "", s)
-    s = re.sub('UUID\(([^)]+)\)', r'\1', s)
+    s = re.sub(r'UUID\(([^)]+)\)', r'\1', s)
     s = re.sub('False', 'false', s)
     s = re.sub('True', 'true', s)
     s = re.sub(r'(ba)=([^[][^ ]*) ', r'\1=[\2] ', s)
@@ -189,7 +190,7 @@ def print_idl(idl, step):
     n = 0
     if "simple" in idl.tables:
         simple = idl.tables["simple"].rows
-        for row in six.itervalues(simple):
+        for row in simple.values():
             s = "%03d: " % step
             s += get_simple_table_printable_row(row)
             print(s)
@@ -197,7 +198,7 @@ def print_idl(idl, step):
 
     if "simple2" in idl.tables:
         simple2 = idl.tables["simple2"].rows
-        for row in six.itervalues(simple2):
+        for row in simple2.values():
             s = "%03d: " % step
             s += get_simple2_table_printable_row(row)
             print(s)
@@ -205,7 +206,7 @@ def print_idl(idl, step):
 
     if "simple3" in idl.tables:
         simple3 = idl.tables["simple3"].rows
-        for row in six.itervalues(simple3):
+        for row in simple3.values():
             s = "%03d: " % step
             s += get_simple3_table_printable_row(row)
             print(s)
@@ -213,7 +214,7 @@ def print_idl(idl, step):
 
     if "link1" in idl.tables:
         l1 = idl.tables["link1"].rows
-        for row in six.itervalues(l1):
+        for row in l1.values():
             s = ["%03d: i=%s k=" % (step, row.i)]
             if hasattr(row, "k") and row.k:
                 s.append(str(row.k.i))
@@ -230,7 +231,7 @@ def print_idl(idl, step):
 
     if "link2" in idl.tables:
         l2 = idl.tables["link2"].rows
-        for row in six.itervalues(l2):
+        for row in l2.values():
             s = ["%03d:" % step]
             s.append(" i=%s l1=" % row.i)
             if hasattr(row, "l1") and row.l1:
@@ -240,13 +241,23 @@ def print_idl(idl, step):
             print(''.join(s))
             n += 1
 
+    if "singleton" in idl.tables:
+        sng = idl.tables["singleton"].rows
+        for row in sng.values():
+            s = ["%03d:" % step]
+            s.append(" name=%s" % row.name)
+            if hasattr(row, "uuid"):
+                s.append(" uuid=%s" % row.uuid)
+            print(''.join(s))
+            n += 1
+
     if not n:
         print("%03d: empty" % step)
     sys.stdout.flush()
 
 
 def substitute_uuids(json, symtab):
-    if isinstance(json, six.string_types):
+    if isinstance(json, str):
         symbol = symtab.get(json)
         if symbol:
             return str(symbol)
@@ -254,14 +265,14 @@ def substitute_uuids(json, symtab):
         return [substitute_uuids(element, symtab) for element in json]
     elif type(json) == dict:
         d = {}
-        for key, value in six.iteritems(json):
+        for key, value in json.items():
             d[key] = substitute_uuids(value, symtab)
         return d
     return json
 
 
 def parse_uuids(json, symtab):
-    if (isinstance(json, six.string_types)
+    if (isinstance(json, str)
             and ovs.ovsuuid.is_valid_string(json)):
         name = "#%d#" % len(symtab)
         sys.stderr.write("%s = %s\n" % (name, json))
@@ -270,29 +281,26 @@ def parse_uuids(json, symtab):
         for element in json:
             parse_uuids(element, symtab)
     elif type(json) == dict:
-        for value in six.itervalues(json):
+        for value in json.values():
             parse_uuids(value, symtab)
 
 
 def idltest_find_simple(idl, i):
-    for row in six.itervalues(idl.tables["simple"].rows):
+    for row in idl.tables["simple"].rows.values():
         if row.i == i:
             return row
     return None
 
 
 def idltest_find_simple2(idl, i):
-    for row in six.itervalues(idl.tables["simple2"].rows):
+    for row in idl.tables["simple2"].rows.values():
         if row.name == i:
             return row
     return None
 
 
 def idltest_find_simple3(idl, i):
-    for row in six.itervalues(idl.tables["simple3"].rows):
-        if row.name == i:
-            return row
-    return None
+    return next(idl.index_equal("simple3", "simple3_by_name", i), None)
 
 
 def idl_set(idl, commands, step):
@@ -334,12 +342,9 @@ def idl_set(idl, commands, step):
             if args[1] == "b":
                 s.b = args[2] == "1"
             elif args[1] == "s":
-                if six.PY2:
-                    s.s = args[2].decode('utf-8')
-                else:
-                    s.s = args[2].encode(sys.getfilesystemencoding(),
-                                         'surrogateescape') \
-                                 .decode('utf-8', 'replace')
+                s.s = args[2].encode(sys.getfilesystemencoding(),
+                                     'surrogateescape') \
+                             .decode('utf-8', 'replace')
             elif args[1] == "u":
                 s.u = uuid.UUID(args[2])
             elif args[1] == "r":
@@ -555,6 +560,11 @@ def do_idl(schema_file, remote, *commands):
     track_notify = False
 
     if remote.startswith("ssl:"):
+        if len(commands) < 3:
+            sys.stderr.write("SSL connection requires private key, "
+                             "certificate for private key, and peer CA "
+                             "certificate as arguments\n")
+            sys.exit(1)
         ovs.stream.Stream.ssl_set_private_key_file(commands[0])
         ovs.stream.Stream.ssl_set_certificate_file(commands[1])
         ovs.stream.Stream.ssl_set_ca_cert_file(commands[2])
@@ -579,11 +589,20 @@ def do_idl(schema_file, remote, *commands):
     else:
         schema_helper.register_all()
     idl = ovs.db.idl.Idl(remote, schema_helper)
+    if "simple3" in idl.tables:
+        idl.index_create("simple3", "simple3_by_name")
 
     if commands:
-        error, stream = ovs.stream.Stream.open_block(
-            ovs.stream.Stream.open(remote))
-        if error:
+        remotes = remote.split(',')
+        stream = None
+        for r in remotes:
+            error, stream = ovs.stream.Stream.open_block(
+                ovs.stream.Stream.open(r), 2000)
+            if not error and stream:
+                break
+            stream = None
+
+        if not stream:
             sys.stderr.write("failed to connect to \"%s\"" % remote)
             sys.exit(1)
         rpc = ovs.jsonrpc.Connection(stream)
@@ -651,18 +670,18 @@ def do_idl(schema_file, remote, *commands):
             step += 1
         else:
             json = ovs.json.from_string(command)
-            if isinstance(json, six.string_types):
+            if isinstance(json, str):
                 sys.stderr.write("\"%s\": %s\n" % (command, json))
                 sys.exit(1)
             json = substitute_uuids(json, symtab)
             request = ovs.jsonrpc.Message.create_request("transact", json)
             error, reply = rpc.transact_block(request)
             if error:
-                sys.stderr.write("jsonrpc transaction failed: %s"
+                sys.stderr.write("jsonrpc transaction failed: %s\n"
                                  % os.strerror(error))
                 sys.exit(1)
             elif reply.error is not None:
-                sys.stderr.write("jsonrpc transaction failed: %s"
+                sys.stderr.write("jsonrpc transaction failed: %s\n"
                                  % reply.error)
                 sys.exit(1)
 
@@ -704,18 +723,18 @@ def do_idl_passive(schema_file, remote, *commands):
 
     for command in commands:
         json = ovs.json.from_string(command)
-        if isinstance(json, six.string_types):
+        if isinstance(json, str):
             sys.stderr.write("\"%s\": %s\n" % (command, json))
             sys.exit(1)
         json = substitute_uuids(json, symtab)
         request = ovs.jsonrpc.Message.create_request("transact", json)
         error, reply = rpc.transact_block(request)
         if error:
-            sys.stderr.write("jsonrpc transaction failed: %s"
+            sys.stderr.write("jsonrpc transaction failed: %s\n"
                              % os.strerror(error))
             sys.exit(1)
         elif reply.error is not None:
-            sys.stderr.write("jsonrpc transaction failed: %s"
+            sys.stderr.write("jsonrpc transaction failed: %s\n"
                              % reply.error)
             sys.exit(1)
 
@@ -732,6 +751,70 @@ def do_idl_passive(schema_file, remote, *commands):
     print("%03d: done" % step)
 
 
+def do_idl_cluster(schema_file, remote, pid, *commands):
+    schema_helper = ovs.db.idl.SchemaHelper(schema_file)
+
+    if remote.startswith("ssl:"):
+        if len(commands) < 3:
+            sys.stderr.write("SSL connection requires private key, "
+                             "certificate for private key, and peer CA "
+                             "certificate as arguments\n")
+            sys.exit(1)
+        ovs.stream.Stream.ssl_set_private_key_file(commands[0])
+        ovs.stream.Stream.ssl_set_certificate_file(commands[1])
+        ovs.stream.Stream.ssl_set_ca_cert_file(commands[2])
+        commands = commands[3:]
+
+    schema_helper.register_all()
+    idl = ovs.db.idl.Idl(remote, schema_helper)
+
+    step = 0
+    seqno = 0
+    commands = list(commands)
+    for command in commands:
+        if command.startswith("+"):
+            # The previous transaction didn't change anything.
+            command = command[1:]
+        else:
+            # Wait for update.
+            while idl.change_seqno == seqno and not idl.run():
+                poller = ovs.poller.Poller()
+                idl.wait(poller)
+                poller.block()
+            step += 1
+
+        seqno = idl.change_seqno
+
+        if command == "reconnect":
+            print("%03d: reconnect" % step)
+            sys.stdout.flush()
+            step += 1
+            idl.force_reconnect()
+        elif command == "remote":
+            print("%03d: %s" % (step, idl.session_name()))
+            sys.stdout.flush()
+            step += 1
+        elif command == "remotestop":
+            r = idl.session_name()
+            remotes = remote.split(',')
+            i = remotes.index(r)
+            pids = pid.split(',')
+            command = None
+            try:
+                command = "kill %s" % pids[i]
+            except ValueError as error:
+                sys.stderr.write("Cannot find pid of remote: %s\n"
+                                 % os.strerror(error))
+                sys.exit(1)
+            os.popen(command)
+            print("%03d: stop %s" % (step, pids[i]))
+            sys.stdout.flush()
+            step += 1
+
+    idl.close()
+    print("%03d: done" % step)
+
+
 def usage():
     print("""\
 %(program_name)s: test utility for Open vSwitch database Python bindings
@@ -801,6 +884,7 @@ def main(argv):
         sys.stderr.write("%s: %s\n" % (ovs.util.PROGRAM_NAME, geo.msg))
         sys.exit(1)
 
+    timeout = None
     for key, value in options:
         if key in ['-h', '--help']:
             usage()
@@ -812,10 +896,11 @@ def main(argv):
             except TypeError:
                 raise error.Error("value %s on -t or --timeout is not at "
                                   "least 1" % value)
-            signal_alarm(timeout)
         else:
             sys.exit(0)
 
+    signal_alarm(timeout)
+
     if not args:
         sys.stderr.write("%s: missing command argument "
                          "(use --help for help)\n" % ovs.util.PROGRAM_NAME)
@@ -833,7 +918,8 @@ def main(argv):
                 "parse-table": (do_parse_table, (2, 3)),
                 "parse-schema": (do_parse_schema, 1),
                 "idl": (do_idl, (2,)),
-                "idl_passive": (do_idl_passive, (2,))}
+                "idl_passive": (do_idl_passive, (2,)),
+                "idl-cluster": (do_idl_cluster, (3,))}
 
     command_name = args[0]
     args = args[1:]
@@ -849,7 +935,7 @@ def main(argv):
             sys.stderr.write("%s: \"%s\" requires at least %d arguments but "
                              "only %d provided\n"
                              % (ovs.util.PROGRAM_NAME, command_name,
-                                n_args, len(args)))
+                                n_args[0], len(args)))
             sys.exit(1)
     elif type(n_args) == int:
         if len(args) != n_args:
